//==========================================================================
// Draw the results of a Gravitational Lens
//
// Author : W. Dhondt
// V01.00
// Created:	2008/06/29
// http://www.kwakkelflap.com/gravlens.html
// Support@kwakkelflap.com
//
//==========================================================================

#include <hpgcc49.h>
#include <hpgraphics.h>

#define G 0.00000000006670
#define PI 3.1415927
#define c 3 * pow(10, 8)
#define arcsec_rad 206265
#define m_in_parsec 3.09 * pow(10, 16)
#define MAX_SCREEN_SIZE 80

BOOL bMethod = TRUE; // TRUE = Plummer model, FALSE = Isotherm Sphere
BOOL bScale = FALSE; // TRUE = logarithmic, FALSE = linear

double MassSource = 5000000000;	// Mass Source (solor masses)
double MassLens = 70000000000;		// Mass Lens (solor masses) (Only needed for Plummer model)
double DistObserverLens = 120;		// Distance observer - lens (Mpc)
double DistObserverSource = 7823;	// Distance observer - source (Mpc)
double RCSource = 1.0;				// kpc
double RCLens = 1.0;				// kpc
double VH = 200;					// km/s (Isotherm Sphere)
double DeltaX = 0.2;				// X Position of source to center (arcsec)
double DeltaY = 0.3;				// Y Position of source to center (arcsec)
double viewangle = 4.0;			// arcsec
int VAL = 0;	// MAX Number of pixels we can draw on the screen

double MASS(double MassLens, double b, double RCLensC, double VHC)
{
	if (bMethod)
	{
		double RetValM = MassLens * pow(b,2) / (pow(b,2) + pow(RCLensC,2));
		return RetValM;
	}
	else
	{
		double RetValM = pow(VHC, 2) * PI * (sqrt(pow(b,2) + pow(RCLensC,2)) - RCLensC) / (2 * G);
		return RetValM;
	}
}

double MU(double DistObserverLens, double DistObserverSource, double DistSourceLens, 
		  double RCLens, double VH, double Theta, double M, double Beta)
{
	if (bMethod)
	{
		double RetValMU = fabs((Theta/Beta) * (1/(1-((DistSourceLens * DistObserverSource * 4 * M * G * (pow(RCLens, 2) - pow(Theta, 2) * pow(DistObserverLens, 2)))/(DistObserverSource * pow(c, 2) * pow((pow(Theta, 2) * pow(DistObserverLens, 2) + pow(RCLens, 2)), 2))))));
		return RetValMU;
	}
	else
	{
		double RetValMU = fabs ((Theta/Beta) * (1/(1-(((2 * pow(VH, 2) * PI * DistSourceLens)/(pow(c, 2) * DistObserverLens * DistObserverSource)) * ((RCLens/pow(Theta, 2))- pow(RCLens, 2)/(pow(Theta, 2) * sqrt((pow(Theta, 2) * pow(DistObserverLens, 2)) + pow(RCLens, 2))))))));
		return RetValMU;
	}
}

int EINSTEINRAD(double MassLens, double DistSourceLens, double DistObserverLens, double DistObserverSource, double RCLens)
{
	double RetTheta = 0;
	if (bMethod)
	{
		RetTheta = sqrt((4 * G * MassLens * DistSourceLens)/(DistObserverLens * pow(c, 2) * DistObserverSource) - (pow(RCLens, 2)/pow(DistObserverLens, 2)));	
		RetTheta *= arcsec_rad;
		RetTheta /= viewangle;
	}
	else
	{
		RetTheta = sqrt((4 * pow(VH, 4) * pow(PI, 2) * pow(DistSourceLens, 2) / (pow(c, 4) * pow(DistObserverSource, 2))) - 4 * pow(VH, 2) * PI * DistSourceLens * RCLens /(DistObserverLens * DistObserverSource * pow(c, 2)));
		RetTheta *= arcsec_rad;
		RetTheta /= viewangle;
	}
	
	int Retra = (int) (RetTheta * VAL);
	return Retra;
}

int main()
{
	int x, y;
	float BigI = 0;
	static float Ifact = 0;
	static int r = 0;
	// Make a matrix of VAL by VAL pixels of greyscale values
	static float I[MAX_SCREEN_SIZE][MAX_SCREEN_SIZE];
	unsigned char color[] = {HPG_COLOR_WHITE, HPG_COLOR_GRAY_1, HPG_COLOR_GRAY_2, HPG_COLOR_GRAY_3, HPG_COLOR_GRAY_4, HPG_COLOR_GRAY_5, HPG_COLOR_GRAY_6, HPG_COLOR_GRAY_7, HPG_COLOR_GRAY_8, HPG_COLOR_GRAY_9, HPG_COLOR_GRAY_10, HPG_COLOR_GRAY_11, HPG_COLOR_GRAY_12, HPG_COLOR_GRAY_13, HPG_COLOR_GRAY_14, HPG_COLOR_BLACK};
	
	unsigned rpl_stack_bias = sat_stack_init(); // Initialize RPL stack
	
	viewangle = sat_pop_real();  	
	if (viewangle == 0.0)
	{
		viewangle = 4.0;
	}
	
	hpg_set_mode_gray16(1); //enter 16 colour, double buffered mode
	hpg_clear_on(hpg_stdscreen);
	hpg_flip();
	hpg_clear_on(hpg_stdscreen);
	
	VAL = min(hpg_get_height(hpg_stdscreen), MAX_SCREEN_SIZE) - 2; // Get the max screen size

    //full speed ahead
    sys_slowOff();

    //This is the actual calculation
	DistObserverLens = DistObserverLens * m_in_parsec *1000000;
	DistObserverSource = DistObserverSource * m_in_parsec *1000000;
	double DistSourceLens = DistObserverSource - DistObserverLens;
	RCLens = RCLens * m_in_parsec *1000;
	RCSource = RCSource * m_in_parsec *1000;
	VH = VH * 1000;
			
	MassSource = MassSource * 1.99 * pow(10, 30);
	MassLens = MassLens * 1.99 * pow(10, 30);
	
	for (x = 0; x < VAL; x++)
	{
		for (y = 0; y < VAL; y++)
		{
					
			double ThetaX = x - (VAL / 2);
			ThetaX /= VAL;
			ThetaX *= viewangle;
			
			double ThetaY = -(y - (VAL / 2));
			ThetaY /= VAL;
			ThetaY *= viewangle;

			double Theta = sqrt(pow(ThetaX, 2) + pow(ThetaY, 2));
									
			double b = Theta * DistObserverLens;
			b /= arcsec_rad;

			
			double bSX = DeltaX * DistObserverSource; 
			bSX /= arcsec_rad;
			double bSY = DeltaY * DistObserverSource;
			bSY /= arcsec_rad;
			
			double ML = MASS(MassLens, b, RCLens, VH);
			
			double Beta = (Theta / arcsec_rad) - ((4 * G * ML * DistSourceLens) / (DistObserverSource * b * pow(c, 2)));
						
			double RX = Beta * DistObserverSource;
			RX *= ThetaX / Theta;
			double RY = Beta * DistObserverSource;
			RY *= ThetaY / Theta;
			double RP = sqrt(pow((RX - bSX), 2) + pow((RY - bSY), 2));

			I[x][y] = (float) (MU(DistObserverLens, DistObserverSource, DistSourceLens, RCLens, VH, Theta, MassLens, Beta) * (MassSource * pow(RCSource, 2)) / (PI * pow((pow(RCSource, 2) + pow(RP, 2)), 2)));
			if (bScale)
			{
				I[x][y] = (float) (log10(I[x][y]));
				if ((I[x][y] > BigI))
				{
					BigI = I[x][y];
				}
			}
			
			if(keyb_isON())
            {
				//back to slow
				sys_slowOn();
				sat_stack_exit(rpl_stack_bias); //close it
		
				return(0);
			}
		}
	}
	
	r = EINSTEINRAD(MassLens, DistSourceLens, DistObserverLens, DistObserverSource, RCLens);
	if (r > VAL/2)
	{
		char errorstring[256];
		
		beep();
		sprintf(errorstring, "Einstein r error: %d", r);
		sat_stack_push_string(errorstring);
		
		//back to slow
		sys_slowOn();
		sat_stack_exit(rpl_stack_bias); //close it
		
		return(0);
	}
	
	if (!bScale)
	{
		int t;
		for (t = -1; t < 1; t++)
		{
			if (BigI < I[(VAL / 2) + r + t][(VAL / 2)])
			{
				BigI = I[(VAL / 2) + r + t][(VAL / 2)];
			}
		}
	}
	
	 //back to slow
    sys_slowOn();
	
	//optional notify/pause for cancel
    beep();
	
	// Draw all pixels on screen
	Ifact = 255 / BigI;
	for (x = 0; x < VAL; x++)
	{
		for (y = 0; y < VAL; y++)
		{
			// Set correct calculated Intensity
			int col = (int) (I[x][y] * Ifact);
			if (col > 255)
			{
				hpg_set_color(hpg_stdscreen, HPG_COLOR_BLACK);
				hpg_draw_pixel_on(hpg_stdscreen, x, y);
			}
			else if (col < 0)
			{
				hpg_set_color(hpg_stdscreen, HPG_COLOR_WHITE);
				hpg_draw_pixel_on(hpg_stdscreen, x, y);
			}
			else
			{
				hpg_set_color(hpg_stdscreen, color[col/16]);
				hpg_draw_pixel_on(hpg_stdscreen, x, y);
			}
		}
	}
	
	hpg_flip(); //updates the screen with the new image (needed in double-buffered mode)
    WAIT_CANCEL
	
	char resultstring[256];
	sprintf(resultstring, "Einstein r %d", r);
	sat_stack_push_string(resultstring);
	
	sat_stack_exit(rpl_stack_bias); //close it

    return(0);
}
